#include <stdio.h>
#include <string.h>
#include <rpc/rpc.h>
#include <unistd.h>
#include <sys/param.h>
#include "_bootUp.h"
///#include "loadTable.c"
#include "../commonIncludes/Table.c"
//#include "../commonIncludes/generalRoutine.c"
#include "../commonIncludes/_defn.h"
#include "../commonIncludes/Coordinator.h"

void passArgumentsToLocalCoordinator(char ipAddr[]); 
    /** This boot strapping procedure will initialize the data which is global for
    *  remote procedures.
	*/

	/* Specifies the time allowed for response to obtain after the broadcast is done */
static struct timeval TIMEOUT = { 25, 0 }; 

NodeArr     toBePassed;

	/* receives replies of broadcast */
NodeArr     tmpArr; 

	/* The value received when some server responds to the broadcast is stored in this variable */
int    retVal;

Bool  replySubsFlag = true;


/**
 * This function is called by this bootUp procedure when it receives the response to its 
 * broadcast from other co-ordinators.
 * One of the functions of this functions is to set a flag that indicates that this host
 * is not the first one in the network, (since it has received the reply from some other
 * co-ordinator.
 */ 
// first arg gives result from server, second server id
int replySubscribe(NodeArr* nodeArr, struct sockaddr_in *addr) {

    int     i;
    char    buff[ MAXHOSTNAMELEN ];
    char    ipAddress[212];
    struct  hostent *name;
    struct  in_addr  in;

	printf(" bootUp.c: Entering the method replySubscribe\n");
	printf("bootUP.c: The length of the node array is %d:\n",nodeArr->NodeArr_len);

		/** 
		 * The value of node-id as -3 indicates that the response has been obtained from the 
		 *  server that is not the last node. In that case, we must wait to receive 
		 *  more replies till we get the response from the last node.
		 *  For the responses from the node that are not the last ones, 
		 *  the node-arr received in the argument is dummy array and hence 
		 *  it can be ignored.
		 */ 
    if(nodeArr->NodeArr_val[0].NodeId == -3)   { 
		
			//The return value of 0 indicates that the host must wait for more replies.
			return WAIT_FOR_MORE_REPLIES; 
	}
	
		/** 
		 * Else it means that response has been obtained from a node that is the last.
		 * The node-arr received as the argument contains the enitire global-node array
		 * and its entries can be copied onto some local node-arr.
		 */
    else {

        toBePassed.NodeArr_len = nodeArr->NodeArr_len;

        toBePassed.NodeArr_val = (Node *) malloc ( 
							toBePassed.NodeArr_len * sizeof(Node));

			/**
			 * Copy the list and information about the nodes received into some local array.
			 */ 
        for(i = 0; i < nodeArr->NodeArr_len; i++) {
            if (nodeArr->NodeArr_val[i].NodeId == -1) {
                toBePassed.NodeArr_val[i].NodeId = -1;
                break;
            }
            Copy(&(toBePassed.NodeArr_val[i]), &(nodeArr->NodeArr_val[i]));
        }
		
			//Set the flag to True to indicate that the host-name is not the first node in the
			//system.
        replySubsFlag = true;

			#ifdef PRINT
				PrintInLogFile( " returning replySub.. \n");
			#endif
				printf( "bootUp.c:  returning replySub.. \n");

			// The return value of 1 indicates that the host should no longer wait for 
			// further responses.
        return DO_NOT_WAIT_FOR_MORE_REPLIES;
    }
}

/** This function method is called when the host machine is the first ARC node in the network.
 *  The purpose of this method is to initialize a node array consisting of two nodes  :
 *  First node-entry contains all the information about the host-machine and its node-id is
 *  "zero".
 *  Second node-entry is similar to the first-node entry expect that its node-id field is -1.
*/ 
void* initializeFirstNode( StrArr* strArr) {

    Node node;

		//Create a node whose id is "zero" and that contains all the information about the 
		//host machine.
    CreateNode(&node, 0, strArr->StrArr_val[0].String_val,
                            strArr->StrArr_val[1].String_val, 
                            (strArr->StrArr_val[2].String_val), 
                            strArr->StrArr_val[3].String_val,
                            atoi(strArr->StrArr_val[4].String_val), 
                            atol(strArr->StrArr_val[5].String_val),
                            atol(strArr->StrArr_val[6].String_val)
                            );
	
    toBePassed.NodeArr_len = 2;
    toBePassed.NodeArr_val = (Node *) malloc (2 * sizeof(Node));

		//Copy the newly created node in the node-array.
    Copy(&(toBePassed.NodeArr_val[0]), &node);

		//Create second node whose id is -1 and this also contains all the information
		//about the host machine.
    CreateNode(&node, -1, strArr->StrArr_val[0].String_val,
                            strArr->StrArr_val[1].String_val, 
                            strArr->StrArr_val[2].String_val, 
                            strArr->StrArr_val[3].String_val,
                            atoi(strArr->StrArr_val[4].String_val), 
                            atol(strArr->StrArr_val[5].String_val),
                            atol(strArr->StrArr_val[6].String_val)
                            );
		//Copy the second node in the second location in the node array.
    Copy(&(toBePassed.NodeArr_val[1]), &node);

	#ifdef PRINT
		PrintInLogFile("Yes at right position %d  \n", toBePassed.NodeArr_val[0].NodeId);
	#endif
}


// if the first argument is the string 'first' then it is first coordinator
// else the process checks whether it is first or not by broadcasting the request and a delay
// is involved in this case, i.e. the delay involved with the time-out.

int main(int argc, char* argv[]) {

     CLIENT           *cl;
     StrArr           *strArr;
     char             buff[MAXHOSTNAMELEN];
     char             ipAddr[50];
     int              *lresult;
     int              i;
     char			  tmp[100];

	/** Get the name of the host machine in the first argument (buff) 
	*  and the ip-address in the second argument (ipAddr) 
	*/

    GetHostNameAndIPadd( buff, ipAddr); 

    strArr = loadTable( NULL); // get all details of current machine
							// stored in a local file called "/kernel/src/node_inf_file".

    for( i = 0; i < strArr->StrArr_len; i++) {

        printf("%s\n", strArr->StrArr_val[i].String_val);
    }

    #ifdef PRINT
    	PrintInLogFile("Just before broadcast"); 
    #endif

		/** whether this is the first node is decided by the command line
		 * argument. We can also decide whether it is the first node  by
		 * broadcasting and asking if any other nodes are present. But this
		 * takes a long time and there are some other problems so for the
		 * time being we decide first node by a command line argument. for
		 * starting the first server give command as boot first.
		 */

	if(argc > 1) {

		if( strcmp(argv[1] , FIRST_NODE_STRING) == 0 ) { 
			
			replySubsFlag = false;  // you know that you are first one
		}
	}
	
	/** If the command line argument does not tell that it is the first node then 
	 *  broadcast the message to find the presence of other ARC servers in the network.
	*/
  	if (replySubsFlag == true)   {

		replySubsFlag = false;

		clnt_broadcast(COORDINATOR, INITIALIZE , BROADCAST_SUBSCRIBE ,
					xdr_StrArr, strArr, xdr_NodeArr, &tmpArr, replySubscribe);

		#ifdef PRINT
			PrintInLogFile("Just after broadcast");     
		#endif
		printf("Just after broadcast\n");
	}
	
		// if you remain to be first, you will receive no replies.
    if (replySubsFlag == false) {
		printf("This is the first node \n");
        initializeFirstNode( strArr);
    }
  
    printf(" Just starting the server --- %s \n",tmp);

		// Start the server.....
    if ( system( "server&" ) != 0 ) {
       error("Major problem can't start the server");
       return 5;
    }

    printf(" Server STARTED ..............\n");

		//Wait for some time after the server has been started to give server some time 
		//to do the initialization.
    sleep(5);

	printf("\n node array  :: length = %d \n", toBePassed.NodeArr_len);

	for(i=0;i< toBePassed.NodeArr_len;i++) {
		printf("--- \n Node Id = %d \n",
		toBePassed.NodeArr_val[i].NodeId);

		printf("Node name = %s \n",
		toBePassed.NodeArr_val[i].Name.String_val);
	}

	passArgumentsToLocalCoordinator(ipAddr);

    clnt_destroy(cl); //Destroy the client handle.
}

void passArgumentsToLocalCoordinator(char ipAddr[]) {
		
	CLIENT *cl;
		//Create a client to the local co-ordinator.
    if ((cl = clnt_create( ipAddr, COORDINATOR, INITIALIZE,"tcp")) == NULL) {
			//If failed then
        clnt_pcreateerror(ipAddr);
        exit( CLIENT_CONNECTION_FAIL );
    }

		// this RPC call goes to the local server
    if (clnt_call(cl, BOOT_UP, xdr_NodeArr, &( toBePassed),  
                    xdr_int, &retVal, TIMEOUT) != RPC_SUCCESS) {

		//If the RPC call to the local server has failed then

        clnt_perror(cl, ipAddr);
        exit(CLIENT_CONNECTION_FAIL);
    }
}
