#include <stdio.h>
#include <time.h>
#include <string.h>
#include <stdarg.h>
#include <netdb.h>
#include <rpc/rpc.h>
#include <sys/param.h>
#include <sys/time.h>
#include <unistd.h>
#include "../commonIncludes/_defn.h"
#include "../commonIncludes/Coordinator.h"
#include "../commonIncludes/generalRoutine.h"

void error(char* fmt, ...) {
    va_list args;

    va_start(args, fmt);
    fprintf(stderr, "error: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr, "\n");
    va_end(args);
}

FILE* efopen(char* file, char* mode) {
    FILE* fptr;

    if( file == NULL) {
	    file = (char*)malloc(sizeof(char) * 20);
	    //strcpy(file,"LOGFILE");
	    strcpy(file,LOGFILE);
    }
    
    if ((fptr = fopen(file, mode)) == NULL)
    {
        error("fopen : can't open file %s", file);
        exit(1);
    }
    return fptr;
}

//void PrintInLogFile(char* fmt, ...) {
void PrintInLogFile(char* fmt, ...) {
    static FILE* logFilePtr;
    static int   count = 1;
    va_list      args;

    if ( count == 1 ) {
        //logFilePtr = efopen("LOGFILE", "a+");
        logFilePtr = efopen(LOGFILE, "a+");
    }

    va_start(args, fmt);
    fprintf(logFilePtr, "%d : ", count++);
    vfprintf(logFilePtr, fmt, args);
    fprintf(logFilePtr, "\n");
    va_end(args);
    fflush(logFilePtr);
}

void PrintInResult(char* fmt, ...) {
    static FILE* ResultPtr;
    static int   count = 1;
    va_list      args;

    if ( count == 1 ) {
        ResultPtr = efopen(RESFILE, "w+");
    }

    va_start(args, fmt);
    fprintf(ResultPtr, "%d : ", count++);
    vfprintf(ResultPtr, fmt, args);
    fprintf(ResultPtr, "\n");
    va_end(args);
    fflush(ResultPtr);
}

void* ecalloc(size_t num, size_t dataSize) {
    void* mem;

    if ((mem = (void*) calloc(num, dataSize)) == NULL)
    {
        error("calloc failed : can't allocate memory");
        return NULL;
    }
    return mem;
}

char* intToStr(int num) {
    int  fr = num / 10;
    int  lastNum = num - fr * 10;
    int  c = lastNum + '0' - 0;
    char *b;

    b = (char *) malloc( 2 * sizeof(char));
    b[0] = c;
    b[1] = '\0';

    if (fr)
    {
        return ((char*) strcat(intToStr(fr), b));
    }
    else return(b);
}


int GetHostNameAndIPadd(char* server, char* ipadd) {
    char   buff[ MAXHOSTNAMELEN ];
    struct hostent *name;
    struct in_addr  in;
    int             len;

    if (gethostname(buff, MAXHOSTNAMELEN)) {
       error("main: gethostname: couldn't get the name of host");
       error("main: Sorry can't boot up");
       return -1;
    }
		
    name = (struct hostent*) malloc(sizeof(struct hostent));
    name = gethostbyname(buff);

    memcpy ((void *) &in, (void *)name->h_addr,(size_t) name->h_length);
     

    if (server != NULL) {
        len = 0;
        while((buff[len] != '\0') &&( buff[len] != '.') && ( buff[len] != '@') ) {
            server[len] = buff[len];
            len++;
        }
        server[len] = '\0';
    }

    if (ipadd != NULL) {
        strcpy(ipadd, (char*) inet_ntoa(in));
        ipadd[strlen(ipadd) + 1] = '\0';
    }
	printf("In general routine : Got the IP as = %s \n",ipadd);

    return 1;
}

void HandleNodeFailure( int id) {
    CLIENT* cl;
    char    server[MAXHOSTNAMELEN];
    struct timeval totaltimeout;

    if (gethostname(server, MAXHOSTNAMELEN)) {
        error("main: gethostname: couldn't get the name of host");
    }
                                    
    if ((cl = clnt_create( server, COORDINATOR, NODE_FAILURE,"tcp")) 
            == NULL) {
         clnt_pcreateerror(server);
         exit( CLIENT_CONNECTION_FAIL );
    }
    timerclear( &totaltimeout);
    
    clnt_call(cl, NODE_FAILED, xdr_int, &id,
          xdr_void, NULL, totaltimeout);
    clnt_destroy(cl);
}

int
live_test( char* IPaddr)
{
    CLIENT* cl;
    
    if ((cl = clnt_create(IPaddr, COORDINATOR, SERVER, "tcp"))
                        == NULL)
    {
        return 0;
    }
    clnt_destroy (cl);
    return 1;
}

/*	Function Added by Satish 
	
	StoreLoad function reads the load of the workstation using w command
	  & also the time when load is taken
	

*/

TimeStamp 
StoreLoad() {
	TimeStamp tmp; 
	FILE *fp;
	char *tok;
	char W[250];
	system("w > /tmp/SMN_W");
	fp = fopen("/tmp/SMN_W","r");
	fscanf(fp,"%[^\n]",W);
	tok = strtok(W," ,:\t");

	PrintInResult("I am in store load ..................\n");

	while((tok = strtok(NULL," ,:\t")) != NULL) {
	   	if(!strcmp(tok,"average"))
			break;
	}
	tmp.load[0] = strtok(NULL," ,:\t"); 
	tmp.load[1] = strtok(NULL," ,\t"); 
	tmp.load[2] = strtok(NULL," \t"); 
	tmp.load[0][5] = '\0';
	tmp.load[1][5] = '\0';
	tmp.load[2][5] = '\0';
	PrintInResult("Load on the machine is ..............:");
	PrintInResult("\n%s\t%s\t%s\n",tmp.load[0],tmp.load[1],tmp.load[2]);	
	tmp.NodeId = -4;
	gettimeofday(&(tmp.Tv),NULL);	
        fclose(fp);
	PrintInResult("\n%s\t%s\t%s\n",tmp.load[0],tmp.load[1],tmp.load[2]);	
	system("rm -f /tmp/SMN_W");
	return(tmp); 
   }


/*	End 	*/


/*
       Main routine to test the functions.
*/


/*
main()
{
  char buff[512];
  char* b;
  efopen("junk", "r");
  error("%s %s %d %s", "I am suppose to", "work well for" , 420, " args");
  PrintInLogFile("Hello . %s..", "kalpesh");
  PrintInLogFile("Hello . %s..", "kalpesh");
  GetHostNameAndIPadd(buff, b);
  strcpy(buff, intToStr(123456));
  printf("%s \n", buff);
  b = intToStr(123456);
  printf("%s \n", b);
}
*/

