#include <stdio.h>
#include <malloc.h>
#include "LockList.h"

/*
#include "generalRoutine.c"
   All functions are checked for all range of value including
   boundary value testing.
*/

void SetLockInfNull(LockList *list) {
    (list->Data).LockId    = -1;
    list->Next             = NULL;
}

/*
     PutInLockList : puts the data at i-th position. Ordering starts
                 from 1 onwards.
                 If position <= 1, it will be put up at first place.
                 If position >= length of string, it will be put up
                 in the end of list.
*/

LockList* PutInLockList(LockList* list, LockInf data, int position) {
	LockList* current;
	LockList* element;
    LockList* tmpPtr;
	int   loopVar = 2;
    LockInf  temp;

    if ((list->Data).LockId == -1) {
        CopyLockInf(&(list->Data), &data);
        list->Next = NULL;
        return list;
    }

	element = (LockList* )malloc(sizeof(LockList));
	element->Next = NULL;

	if(position <= 1) {
        element->Next = list->Next;
        list->Next    = element;

        CopyLockInf( &(element->Data), &(list->Data));
        CopyLockInf( &(list->Data), &data);
    }
	else {
		current       = list;
        CopyLockInf( &(element->Data), &data);

		while ((current->Next) && (loopVar < position)) {
			current = current->Next;
			loopVar++;
		}
        element->Next = current->Next;
        current->Next = element;
	}
    return element;
}

int LockListLength(LockList* list) {
    int len = 0;
    LockList* current = list;
    
    if( list == NULL) {
	return 0;   
    }
 
    if ((current->Data).LockId == -1) {
        return 0;
    }
    
    while (current) {
        current = current->Next;
        len++;
    }
    return len;
}

/*
    DestroyLockList : Destroys the list except the first one.
                  Makes first->Next null
*/

void DestroyLockList(LockList* list) {
	LockList* current = list->Next;
	LockList* prev;

	while(current) {
        prev    = current;
		current = current->Next;
		free(prev);
	}
    list->Next = NULL;
}

/*
    GetLink : Returns the link at position pos.
              if pos <= 0 or > length of list returns NULL;
*/

LockList* GetLockInfLink(LockList* list, int pos) {
	LockList* current = list;
	int   loopVar = 1;

    if ((pos < 1) && (pos > LockListLength(list))) {
        fprintf(stderr, "Error : can't delete data index %d\n", pos);
        return NULL;
    }
    
	while((loopVar < pos) && (current)) {
		current = current->Next;
		loopVar++;
	}
	return current;
}

LockList* GetLockLink(LockList* list, int* pos, int lockId) {
    int loopVar = 0; 

	LockList* current = list;

    if (current == NULL) {
        return NULL;
    }
    
    if ((current->Data).LockId == -1) {
        return NULL;
    }
    
    while (current) {
        loopVar++;
        if ( current->Data.LockId == lockId ) {
            *pos = loopVar;
            return current;
        }
		current = current->Next;
	}
    return ( current);
}
/*
    GetData : return's the data at position pos.
*/

LockInf DeleteLockInfLink(LockList* list,int pos) {
	LockList* current = list;
    LockList* prev    = NULL;
	int   loopVar = 1;
	LockInf  retp;

    if ((pos < 1) || (pos > LockListLength(list))) {
        error("Error : can't delete pos %d \n", pos);
        exit(1);
    }
    
    if (pos == 1) {
        if (  LockListLength(list) > 1) {
            CopyLockInf( &retp, &(list->Data));
            CopyLockInf(&(list->Data), &(list->Next->Data));
            prev = list->Next;
            list->Next = list->Next->Next;
            free(prev);
        }
        else {
               (list->Data).LockId = -1;
        }
    }
    else {
	    while((loopVar < pos) && (current)) {
            prev    = current;
	    	current = current->Next;
		    loopVar++;
	    }

        if (!current) {
            fprintf(stderr, "Error : can't delete data index %d\n", pos);
            exit(1);
        }

	    CopyLockInf(&retp, &(current->Data));
        prev->Next = current->Next;
	    free(current);
    }
	return retp;
}

/*
    PrintLockList :  prints the list, provided data is printable.
*/

void PrintLockList(LockList *list) {
	LockList* current = list;


#ifdef PRINT

	PrintInLogFile("The list of local locks at the time of initialization i.e. the time when depost locks wa called. \n");
    while (current) {
		PrintInLogFile(" %d %d \n", current->Data.LockId, current->Data.Status);
		current = current->Next;
	}

#endif

}

int GetFreeLock(LockList* list) {
    int pos = 0;
    
	LockList* current = list;

    if ((current->Data).LockId == -1) {
        return 2;
    }
    
    while (current) {
        pos++;
        if ( current->Data.Status == 1 ) {
            return pos;
        }
		current = current->Next;
	}
    return (pos + 1);
}

void CopyLockInf(LockInf* node1, LockInf* node2) {
        node1->LockId = node2->LockId;
        node1->Status = node2->Status;
}

