#include <stdio.h>
#include <malloc.h>
#include "LockReqList.h"

/*
#include "generalRoutine.c"
   All functions are checked for all range of value including
   boundary value testing.
*/

void SetLockReqInfNull(LockReqList *list)
{
    if( list == NULL) {
        list = (LockReqList *) malloc(sizeof(LockReqList));
    }
    (list->Data).NodeId = -1;
    (list->Data).LockId = -1;
    list->Next             = NULL;
}

/*
     PutInLockReqList : puts the data at i-th position. Ordering starts
                 from 1 onwards.
                 If position <= 1, it will be put up at first place.
                 If position >= length of string, it will be put up
                 in the end of list.
*/

LockReqList* PutInLockReqList(
						LockReqList* list, 
						LockReqInf data, 
						int position) {

	LockReqList* current;
	LockReqList* element;
    LockReqList* tmpPtr;
	int   loopVar = 2;
    LockReqInf  temp;

    if ((list->Data).LockId == -1) {
        CopyLockReqInf(&(list->Data), &data);
        list->Next = NULL;
        return list;
    }
    
	element = (LockReqList* )malloc(sizeof(LockReqList));
	element->Next = NULL;

	if(position <= 1) {
        element->Next = list->Next;
        list->Next    = element;

        CopyLockReqInf( &(element->Data), &(list->Data));
        CopyLockReqInf( &(list->Data), &data);
    }
	else {
		current       = list;
        CopyLockReqInf( &(element->Data), &data);

		while ((current->Next) && (loopVar < position)) {
			current = current->Next;
			loopVar++;
		}
        element->Next = current->Next;
        current->Next = element;
	}
    return element;
}

int LockReqListLength(LockReqList* list) {
    int len = 0;
    LockReqList* current = list;
    
    if( list == NULL) {
	return 0;   
    }
 
    if ((current->Data).LockId == -1) {
        return 0;
    }
    
    while (current) {
        current = current->Next;
        len++;
    }
    return len;
}

/*
    DestroyLockReqList : Destroys the list except the first one.
                  Makes first->Next null
*/

void DestroyLockReqList(LockReqList* list) {
	LockReqList* current = list->Next;
	LockReqList* prev;

	while(current)
	{
        prev    = current;
		current = current->Next;
		free(prev);
	}
    list->Next = NULL;
}

/*
    GetLink : Returns the link at position pos.
              if pos <= 0 or > length of list returns NULL;
*/

LockReqList* GetLockReqInfLink(LockReqList* list, int pos) {
	LockReqList* current = list;
	int   loopVar = 1;

    if ((pos < 1) && (pos > LockReqListLength(list)))
    {
        fprintf(stderr, "Error : can't delete data index %d\n", pos);
        return NULL;
    }
    
	while((loopVar < pos) && (current))
	{
		current = current->Next;
		loopVar++;
	}
	return current;
}

LockReqList* GetLockReqLink(
					LockReqList* list, 
					int* pos, 
					int nodeId, 
					int lockId, 
					int seqNum) {

    LockReqList* current = list;
    int   loopVar = 1;

    if(list == NULL) {
        return NULL;
    }
 
    if((list->Data).NodeId == -1) {
        return NULL;
    }

    while(current) {
        if ((current->Data.NodeId == nodeId) &&
                (current->Data.LockId == lockId) &&
                ( current->Data.SeqNum == seqNum)) {

            break;
        }
        current = current->Next;
		loopVar++;
    }

    pos = &loopVar;
    return current;
}


/*
    GetData : return's the data at position pos.
*/

LockReqInf
DeleteLockReqInfLink(LockReqList* list,int pos)
{
	LockReqList* current = list;
    LockReqList* prev    = NULL;
	int   loopVar = 1;
	LockReqInf  retp;

    if ((pos < 1) || (pos > LockReqListLength(list)))
    {
        error("Error : can't delete pos %d \n", pos);
        exit(1);
    }
    
    if (pos == 1)
    {
        if (  LockReqListLength(list) > 1)
        {
            CopyLockReqInf( &retp, &(list->Data));
            CopyLockReqInf(&(list->Data), &(list->Next->Data));
            prev = list->Next;
            list->Next = list->Next->Next;
            free(prev);
        }
        else
        {
               (list->Data).NodeId = -1;
        }
    }
    else
    {
	    while((loopVar < pos) && (current))
	    {
            prev    = current;
	    	current = current->Next;
		    loopVar++;
	    }

        if (!current)
        {
            fprintf(stderr, "Error : can't delete data index %d\n", pos);
            exit(1);
        }

	    CopyLockReqInf(&retp, &(current->Data));
        prev->Next = current->Next;
	    free(current);
    }
	return retp;
}

/*
    PrintLockReqList :  prints the list, provided data is printable.
*/

void
PrintLockReqList(LockReqList *list)
{
    int i;

	LockReqList* current = list;


#ifdef PRINT

    while (current) {
		PrintInLogFile(" %d %d %d\n", current->Data.NodeId, 
                current->Data.LockId, current->Data.SeqNum);

        for( i =0; i < current->Data.Len; i++) {
            PrintInLogFile( " %d ", current->Data.ReplyCount[i]); 
        }
        PrintInLogFile( " \n "); 
		current = current->Next;
	}

#endif

}

void CopyLockReqInf(LockReqInf* node1, LockReqInf* node2) {
        int i;
    
        node1->NodeId = node2->NodeId;
        node1->LockId = node2->LockId;
        node1->SeqNum = node2->SeqNum;
        node1->Len    = node2->Len;
        
        node1->ReplyCount =  (int *) malloc(node2->Len * sizeof( int));
         
        for( i =0; i < node1->Len; i++)
        {
            node1->ReplyCount[i] = node2->ReplyCount[i]; 
        }
}

