#include <stdio.h>
#include <malloc.h>
#include "Coordinator.h"

/*
   All functions are checked for all range of value including
   boundary value testing.
*/

void SetNullString(String *str) {
    str->String_len = 0;
    str->String_val = NULL;
}

void SetTableNull(Table *list) {
    (list->NodeInf).NodeId    = -1;
	list->Next                = NULL;

    SetNullString(&((list->NodeInf).Name));
    SetNullString(&((list->NodeInf).IPaddress));
}

void SetNodeNull(Node* node) {
     node->NodeId           = -1;
    (node->Name).String_len = 0;
    (node->Name).String_val = NULL;

    (node->IPaddress).String_len = 0;
    (node->IPaddress).String_val = NULL;

    (node->ArchType).String_len = 0;
    (node->FileSys).String_val = NULL;

    (node->ArchType).String_len = 0;
    (node->FileSys).String_val = NULL;
}

void PutInString(String* str, char* strval) {
    str->String_len = strlen(strval);
    printf("string to be copied is %s\n", strval);
    if (str->String_val != NULL) {
/*
        free(str->String_val);
        str->String_val = NULL;
*/
    }
    str->String_val = (char* )malloc((str->String_len + 1) *  sizeof(char));
    strcpy(str->String_val, strval);
    str->String_val[ str->String_len ] = '\0';
//    printf("In Copy:PutInString:- string value is %s\n", str->String_val );
}

void Copy(Node* node1, Node* node2) {
    node1->NodeId = node2->NodeId;
//    printf("In Copy: Value of second parameter is :%s \n ", node2->Name.String_val );
//    printf("In Copy: length of second parameter is :%d \n ", node2->Name.String_len );
    node2->Name.String_val[ node2->Name.String_len ] = '\0';
    PutInString(&(node1->Name), (node2->Name).String_val);
//    printf("In Copy: Value of first parameter after COPY is :%s \n ", node1->Name.String_val );

    node2->IPaddress.String_val[ node2->IPaddress.String_len ] = '\0';
    PutInString(&(node1->IPaddress), (node2->IPaddress).String_val);

    node2->ArchType.String_val[ node2->ArchType.String_len ] = '\0';
    PutInString(&(node1->ArchType), (node2->ArchType).String_val);
    
    node2->FileSys.String_val[ node2->FileSys.String_len ] = '\0';
    PutInString(&(node1->FileSys), (node2->FileSys).String_val);

    node1->Capacity = node2->Capacity;
    node1->RangeBegin = node2->RangeBegin;
    node1->RangeEnd = node2->RangeEnd;
}

void
Swap(Node *node1, Node *node2)
{
    Node temp;
    SetNodeNull(&temp);

    Copy(&temp, node1);
    Copy(node1, node2);
    Copy(node2, &temp);
}

void
CreateNode(Node* node, int nodeId, char* name,
                   char* ipAdd, char* arch, char* fileSys, 
                   int capacity, u_long num1, u_long num2)
{
    node->NodeId = nodeId;
    PutInString(&(node->Name), name);
    PutInString(&(node->IPaddress), ipAdd);
    PutInString(&(node->ArchType), arch);
    PutInString(&(node->FileSys), fileSys);
    node->Capacity = capacity;
    node->RangeBegin = num1;
    node->RangeEnd = num2;
}

/*
     PutInTable : puts the data at i-th position. Ordering starts
                 from 1 onwards.
                 If position <= 1, it will be put up at first place.
                 If position >= length of string, it will be put up
                 in the end of list.
*/

Table* PutInTable(Table* list, Node* node, int position) {
	Table* current;
	Table* element;
	int    loopVar = 2;

    if ((list->NodeInf).NodeId == -1) {
        if (list->Next != NULL) {
            PrintInLogFile("Error : SetNull not called for Table \n");
            return NULL;
        }
        Copy(&(list->NodeInf), node);
        return list;
    }

	element = (Table* )malloc(sizeof(Table));
    SetTableNull(element);
	element->Next = NULL;

	if(position <= 1) {
        element->Next = list->Next;
        list->Next    = element;

        Copy(&(element->NodeInf), &(list->NodeInf));
        Copy(&(list->NodeInf), node);
    }

	else {
		current       = list;
        Copy(&(element->NodeInf), node);

		while ((current->Next) && (loopVar < position)) {
			current = current->Next;
			loopVar++;
		}
        element->Next = current->Next;
        current->Next = element;
	}
    return element;
}

void CopyTable(TablePtr table1, TablePtr table2) {
	Table* current = table2;
    Table* c2 = table2;

    SetTableNull(table1);

    while (current)
	{
        PutInTable(table1, &(current->NodeInf),
                                 (current->NodeInf).NodeId);
		current = current->Next;
	}
}

int
Length(Table list)
{
    int len = 0;
    Table* current = &list;

    if ((current->NodeInf).NodeId == -1)
    {
        return 0;
    }

    while (current)
    {
        current = current->Next;
        len++;
    }
    return len;
}

/*
    DestroyTable : Destroys the list except the first one.
                   Makes first->Next null
*/

void
DestroyTable(Table* list)
{
	Table* current = list->Next;
	Table* prev;

	while(current)
	{
        prev    = current;
		current = current->Next;
        free((prev->NodeInf).Name.String_val);
        free((prev->NodeInf).IPaddress.String_val);
        free((prev->NodeInf).FileSys.String_val);
		free(prev);
	}
    list->Next = NULL;
}

/*
    GetLink : Returns the link at position pos.
              if pos <= 0 or > length of list returns NULL;
*/

Table*
GetLink(Table* list, int pos)
{
	Table* current = list;
	int   loopVar = 1;

    if (pos <= 0)
    {
        return NULL;
    }

	while((loopVar < pos) && (current))
	{
		current = current->Next;
		loopVar++;
	}
	return current;
}

/*
    GetData : return's the data at position pos.
*/

Node* GetNodeInf(Table* list, int pos) {
    Table*  link = GetLink(list, pos);
    char    buff[200];

    if (link == NULL) {
        sprintf(buff,"Error : GetData invalid index %d", pos); 
        PrintInLogFile(buff);
        exit(1);
    }
	PrintInLogFile(" \n");
    return &(link->NodeInf);
}

/*
    DeleteLink : Deletes the data at position pos.
*/

Node* DeleteLink(Table* list, int pos) {
	Table* current = list;
    Table* prev    = NULL;
	int    loopVar = 1;
	Node*  retp;
    char   buff[200];

    if (pos < 1) {
        sprintf(buff, "Error : can't delete data index %d", pos);
        PrintInLogFile(buff);
	PrintInLogFile(" \n");
        exit(1);
    }
    
    if (pos == 1) {
        if ( Length(*list) > 1) {
            Copy(retp, &(list->NodeInf));
            Copy(&(list->NodeInf), &(list->Next->NodeInf));
            prev       = list->Next;
            list->Next = list->Next->Next;
            free(prev);
        }
        else {
            (list->NodeInf).NodeId = -1;
        }
    }
    else {
	    while((loopVar < pos) && (current)) {
            prev    = current;
	    	current = current->Next;
		    loopVar++;
	    }

        if (!current) {
            sprintf(buff,"Error : can't delete data index %d", pos);
            PrintInLogFile(buff);
			PrintInLogFile(" \n");
            exit(1);
        }

	    Copy(retp, &(current->NodeInf));
        prev->Next = current->Next;
	    free(current);
    }
	return retp;
}

/*
    PrintTable :  prints the list, provided data is printable.
*/

void PrintTable(NodeArr list, int last) {
    char   buff[ 1024 ];
    int     i;
    
#ifdef PRINT

    PrintInLogFile("%d \n",last);
    for( i = 0; i < last; i++) {
        PrintInLogFile(buff, "%d %s %s %s %s %d %x %x \n", 
                 list.NodeArr_val[i].NodeId,
                 list.NodeArr_val[i].Name.String_val,
                 list.NodeArr_val[i].IPaddress.String_val,
                 list.NodeArr_val[i].ArchType.String_val,
                 list.NodeArr_val[i].FileSys.String_val,
                 list.NodeArr_val[i].Capacity,
                 list.NodeArr_val[i].RangeBegin,
                 list.NodeArr_val[i].RangeEnd);
	}

#endif

}

int GetTheLastNodeId(Table* table, Node* node, int* pos) {
    Table* current = table;
    int    max     = -1;
    int    count   = 0;

    SetNodeNull(node);

    while (current)
    {
        count++;
        if (max < (current->NodeInf).NodeId)
        {
             max = current->NodeInf.NodeId;
             if (pos != NULL)
             {
                *pos = count;
             }
             if (node != NULL)
             {
                Copy(node, &(current->NodeInf));
             }
        }
        current = current->Next;
    }
    return max;
}

int
GetNodeId(Table* table, char* name, int* pos)
{
    Table* current = table;
    int    myId    = -1;
    int    count   = 0;
    
    if (pos)
    {
        *pos = -1;
    }

    while (current)
    {
        count++;
        if ( !strcmp(name, (current->NodeInf).Name.String_val))
        {
             myId = current->NodeInf.NodeId;
             if (pos != NULL)
             {
                *pos = count;
             }
             break;
        }
        current = current->Next;
    }
    return myId;
}

int
GetNodeWithId(TablePtr table, int id, Node* node)
{
	Table* current = table;
    int    pos     = 0;

    if ((current->NodeInf).NodeId == -1)
    {
        node = NULL;
        return 0;
    }

	while(current)
	{
        pos++;
        if ((current->NodeInf).NodeId == id)
        {
            node = &(current->NodeInf);
            break;
        }
		current = current->Next;
	}
	return pos;
}

void
DeleteNodeWithId(TablePtr table, int seqNo)
{
    int pos = GetNodeWithId(table, seqNo, NULL);

    DeleteLink(table, pos);
}

void
RemoveFromTable(Table* table, char* name)
{
    int pos;
    int id = GetNodeId(table, name, &pos);
    DeleteLink(table, pos);
}
